# tests/test_log.py

import sys
sys.path.insert(0, "/data/tianzhen/my_packages/cobra-log/src")

from cobra_log import (warning, info, error, log_init, set_trace, critical)


def test_cobra_log():
    try:
        try:
            raise KeyError("This is the first exception")
        except Exception:
            try:
                raise KeyError("This is the second exception")
            except Exception:
                # error("An error occurred during the test.", throw=True)
                raise critical("A critical error occurred during the test1.", throw=None)

    except Exception:
        warning("An error occurred during the test.")
        info("Continuing execution after warning.")
        error("An error occurred during the test.", "")
        # critical("A critical error occurred during the test.")


def test_cobra_log_2():
    try:
        raise KeyError("This is the first exception")

    except Exception as e:
        warning("An error occurred during the test.", e, loc=True)
        info("Continuing execution after warning.")
        info("Continuing execution after warning.", e, outline=True)
        critical("A critical error occurred during the test.")

def test_cobra_log_3():
    try:
        try:
            try:
                1 / 0
            except Exception as e:
                raise error("An error occurred during the test31.", throw=None) from e
        except Exception as ee:
            raise error("An error occurred during the test32.", throw=TimeoutError) from ee
    except Exception:
        critical("A critical error occurred during the test33.", throw=None)


def test_cobra_log_4():
    try:
        try:
            1 / 0
        except ZeroDivisionError as e:
            try:
                int("abc")  # 这里又抛出一个异常
            except ValueError as ve:
                new_exc = RuntimeError("New exception")
                raise new_exc from ve  # 将原始异常作为新异常的 cause
    except Exception as e:
        print("Caught exception:", e)
        print("Cause:", e.__cause__)
        print("Context:", e.__context__)
        print("Suppress Context:", e.__suppress_context__)
        # raise
        error("An error occurred")


if __name__ == "__main__":
    log_init("test_log.log", log_level="debug", use_color=True)
    set_trace(with_border=True, exc_depth=3, tb_depth=3, exc_args_limit=3, min_width=1)
    test_cobra_log()
    print("\n" + "=" * 80 + "\n")
    test_cobra_log_2()
    print("\n" + "=" * 80 + "\n")
    test_cobra_log_3()
    print("\n" + "=" * 80 + "\n")
    test_cobra_log_4()




class CriticalErrorException(Exception):
    """Exception raised for critical errors in the cobra_log module."""


class ErrorException(Exception):
    """Exception raised for errors in the cobra_log module."""


class WarningException(Exception):
    """Exception raised for warnings in the cobra_log module."""



# -*- coding: utf-8 -*-
# Python version: 3.9
# @TianZhen

from __future__ import annotations
import sys
import inspect
from typing import (Any, Optional, Union, Literal)

from . import _core
from ._exceptions import (
    CriticalErrorException,
    ErrorException,
    WarningException
)
from ._utils import (trace_exc, trace_stack, box_lines, _fmt_msg)


_EXC_FMT = r" [%(fileName)s->%(funcName)s(%(lineno)d)]"
_FRAME_GAP = 2


def _fmt_main(prefix: str, msg: str = "", /, loc: Union[bool, int] = True, indent: int = 0, **pattern: Any):
    r"""Format the main message."""
    _msg = _fmt_msg(msg)
    # location information
    if loc is False:
        loc_info = ""
    else:
        loc_level = 0 if loc is True else max(0, int(loc))
        loc_info = trace_stack(loc_level + 3, fmt=_EXC_FMT)
    # main message
    main_msg = _core.cstr(f"{prefix}{loc_info}: {_msg}", **pattern)
    # indent
    if indent > 0:
        main_msg = _core.cstr(" " * indent, main_msg)

    return main_msg, _msg


def _fmt_exc(exc: Any, /, top_indent: int, frame_style: str, indent: int = 0, **pattern: Any):
    r"""Format the exception trace."""
    # exception
    if exc is None:
        exc = sys.exc_info()[1]  # try to get the current exception
    if not exc:
        return "", None

    with_border = bool(_core._TRACE_CONFIG.get("with_border", True))
    # exception message
    exc_msg = trace_exc(
        exc,
        mode=_core._TRACE_CONFIG.get("exc_mode", "context"),
        exc_depth=_core._TRACE_CONFIG.get("exc_depth", -1),
        tb_depth=_core._TRACE_CONFIG.get("tb_depth", -1),
        exc_args_limit=_core._TRACE_CONFIG.get("exc_args_limit", -1),
        indent=3 if with_border else 4
    )
    # box lines
    if with_border:
        exc_msg = box_lines(
            exc_msg.splitlines()[1:],
            top_indent=top_indent + _FRAME_GAP,
            rest_indent=indent,
            frame_style=frame_style,
            **pattern
        )

    return exc_msg, exc


def critical(
    msg: str = "",
    /,
    exc: Optional[Any] = None,
    throw: Union[Optional[Exception], Literal["fatal"]] = "fatal",
    loc: Union[bool, int] = True,
    display: bool = True
):
    r"""
    `CRITICAL-ERROR` or `FATAL-ERROR` level with `log` record (requires :func:`log_init`).

    NOTE: Used when the overall program is not running.

    Parameters
    ----------
        msg : str, default to `""`
            Description message.

        exc : Optional[Any], default to `None`
            An instance of exception.
            - _Exception_: Record the exception traceback and message;
            - `None`: Try to catch the exception in the current context. If no exception is caught, no exception information will be recorded;
            - `""`: No exception information will be recorded.

        throw : Union[Optional[Exception], Literal["fatal"]], default to `"fatal"`
            The type of exception to be output after logging, use `raise` to throw the exception.
            - `"fatal"`: Return a :class:`SystemExit` exception with the formatted message, which will cause the program to `exit` immediately;
            - _Exception_: Return the specified exception with the description message;
            - `None`: Return :class:`CriticalErrorException` exception with the description message.

        loc : Union[bool, int], default to `True`
            Control whether to display file function location information.
            - `True`: Display `CRITICAL` location as stack level `1`;
            - _int_: Display `CRITICAL` location according to stack level. i.e.: `0`: the location where the `CRITICAL` is called; `1`: the location of the parent function that calls the `CRITICAL`.; ...

        display : bool, default to `True`
            Control whether to display the formatted message.

    Returns
    -------
        SystemExit
            :param:`throw` is `"fatal"`.
        Exception
            :param:`throw` is an Exception class.
        CriticalErrorException
            :param:`throw` is `None`.
    """
    # main message
    prefix = "{FATAL-ERROR}" if throw == "fatal" else "CRITICAL-ERROR"
    main_msg, _msg = _fmt_main(prefix, msg, loc=loc, fg="w", bg="lr", styles={"bold"})
    # exception message
    exc_msg, exc = _fmt_exc(exc, top_indent=len(main_msg), frame_style="double", fg="lr", styles={"bold", "blink"})
    # log
    if _core._FILE_HANDLER:
        _core._LOGGER.critical(str(_msg), exc_info=exc, stack_info=True, stacklevel=2)
    # combine
    final_msg = _core.cstr(main_msg, " " * _FRAME_GAP, exc_msg)
    # display
    if display:
        _core.display(final_msg)
    # throw
    if throw == "fatal":
        return SystemExit(final_msg)
    elif inspect.isclass(throw) and issubclass(throw, Exception):
        return throw(_msg)
    else:
        return CriticalErrorException(_msg)


def error(
    msg: str = "",
    /,
    exc: Optional[Any] = None,
    throw: Optional[Exception] = None,
    loc: Union[bool, int] = True,
    display: bool = True
):
    r"""
    `ERROR` level with `log` record (requires :func:`log_init`).

    NOTE: Used when some functions are not running.

    Parameters
    ----------
        msg : str, default to `""`
            Description message.

        exc : Optional[Any], default to `None`
            An instance of exception.
            - _Exception_: Record the exception traceback and message;
            - `None`: Try to catch the exception in the current context. If no exception is caught, no exception information will be recorded;
            - `""`: No exception information will be recorded.

        throw : Optional[Exception], default to `None`
            The type of exception to be output after logging.
            - _Exception_: Return the specified exception with the description message;
            - `None`: Return :class:`ErrorException` exception with the description message.

        loc : Union[bool, int], default to `True`
            Control whether to display file function location information.
            - `True`: Display `ERROR` location as stack level `1`;
            - _int_: Display `ERROR` location according to stack level. i.e.: `0`: the location where the `ERROR` is called; `1`: the location of the parent function that calls the `ERROR`.; ...

        display : bool, default to `True`
            Control whether to display the formatted message.

    Returns
    -------
        Exception
            :param:`throw` is an Exception class.
        ErrorException
            :param:`throw` is `None`.
    """
    # main message
    main_msg, _msg = _fmt_main("ERROR", msg, loc=loc, fg="d", bg="y", styles={"bold"})
    # exception message
    exc_msg, exc = _fmt_exc(exc, top_indent=len(main_msg), frame_style="double", fg="y", styles={"bold", "blink"})
    # log
    if _core._FILE_HANDLER:
        _core._LOGGER.error(str(_msg), exc_info=exc, stack_info=True, stacklevel=2)
    # display
    if display:
        # combine
        final_msg = _core.cstr(main_msg, " " * _FRAME_GAP, exc_msg)
        _core.display(final_msg)
    # throw
    if inspect.isclass(throw) and issubclass(throw, Exception):
        return throw(_msg)
    else:
        return ErrorException(_msg)


def warning(
    msg: str = "",
    /,
    exc: Optional[Any] = None,
    throw: Optional[Exception] = None,
    loc: Union[bool, int] = True,
    dim: bool = False,
    display: bool = True
):
    r"""
    `WARNING` level with `log` record (requires :func:`log_init`).

    NOTE: Used when unexpected events occur, and the program can still run normally.

    Parameters
    ----------
        msg : str, default to `""`
            Description message.

        exc : Optional[Any], default to `None`
            An instance of exception.
            - _Exception_: Record the exception traceback and message;
            - `None`: Try to catch the exception in the current context. If no exception is caught, no exception information will be recorded;
            - `""`: No exception information will be recorded.

        throw : Optional[Exception], default to `None`
            The type of exception to be output after logging.
            - _Exception_: Return the specified exception with the description message;
            - `None`: Return :class:`WarningException` exception with the description message.

        loc : bool, default to `True`
            Control whether to display file function location information.
            - `True`: Display `WARNING` location as stack level `1`;
            - _int_: Display `WARNING` location according to stack level. i.e.: `0`: the location where the `WARNING` is called; `1`: the location of the parent function that calls the `WARNING`.; ...

        dim : bool, default to `False`
            Control whether to dim the `WARNING` message.

        display : bool, default to `True`
            Control whether to display the formatted message.

    Returns
    -------
        Exception
            :param:`throw` is an Exception class.
        WarningException
            :param:`throw` is `None`.
    """
    # main message
    main_msg, _msg = _fmt_main("WARNING", msg, loc=loc, fg="y", styles=None if dim else {"bold"})
    # exception message
    exc_msg, exc = _fmt_exc(exc, top_indent=len(main_msg), frame_style="light", fg="y", styles={"dim"} if dim else {"bold"})
    # log
    if _core._FILE_HANDLER:
        _core._LOGGER.warning(str(_msg), exc_info=exc, stack_info=True, stacklevel=2)
    # display
    if display:
        # combine
        final_msg = _core.cstr(main_msg, " " * _FRAME_GAP, exc_msg)
        _core.display(final_msg)
    # throw
    if inspect.isclass(throw) and issubclass(throw, Exception):
        return throw(_msg)
    else:
        return WarningException(_msg)


def info(
    msg: str = "",
    /,
    exc: Optional[Any] = "",
    indent: int = 0,
    loc: Union[bool, int] = False,
    outline: bool = False,
    display: bool = True
):
    r"""
    `INFO` level with `log` record (requires :func:`log_init`).

    NOTE: Used to record key node information.

    Parameters
    ----------
        msg : str, default to `""`
            Description message.

        exc : Optional[Any], default to `""`
            An instance of exception.
            - _Exception_: Record the exception traceback and message;
            - `None`: Try to catch the exception in the current context. If no exception is caught, no exception information will be recorded;
            - `""`: No exception information will be recorded.

        indent : int, default to `0`
            The indentation of the `INFO`.

        loc : bool, default to `True`
            Control whether to display file function location information.
            - `True`: Display `INFO` location as stack level `1`;
            - _int_: Display `INFO` location according to stack level. i.e.: `0`: the location where the `INFO` is called; `1`: the location of the parent function that calls the `INFO`.; ...

        outline : bool, default to `False`
            Control whether to highlight the `INFO` message.

        display : bool, default to `True`
            Control whether to display the formatted message.

    Returns
    -------
        str
            Formatted message.
    """
    # main message
    if outline:
        main_msg, _msg = _fmt_main("KEY-INFO", msg, loc=loc, indent=indent, fg="lb", styles={"bold"})
    else:
        main_msg, _msg = _fmt_main("INFO", msg, loc=loc, indent=indent, fg="lg", styles={"bold"})
    # exception message
    exc_msg, exc = _fmt_exc(exc, top_indent=len(main_msg), indent=indent, frame_style="light", fg="lb" if outline else "g", styles={"bold"} if outline else None)
    # log
    if _core._FILE_HANDLER:
        _core._LOGGER.info(str(_msg), exc_info=exc, stack_info=False, stacklevel=2)
    # combine
    final_msg = _core.cstr(main_msg, " " * _FRAME_GAP, exc_msg)
    # display
    if display:
        _core.display(final_msg)

    return final_msg


def debug(*args: Any, **kwargs: Any):
    r"""
    `DEBUG` level with `log` record (requires :func:`log_init`).

    NOTE: Used for debugging.
    """
    msg = ""
    for arg in args:
        msg += f"\n{arg}"
    for arg_name, arg_val in kwargs.items():
        msg += f"\n[{arg_name}]: {arg_val}"
    # log
    if _core._FILE_HANDLER and msg:
        _core._LOGGER.debug(msg, stack_info=True, stacklevel=2)



# -*- coding: utf-8 -*-
# Python version: 3.9
# @TianZhen

import inspect
import os
import traceback
from typing import (Any, Tuple, List, Optional, Sequence, Literal)

from ._core import cstr, _TRACE_CONFIG


_TRACE_FMT = r"%(fileName)s->%(funcName)s(%(lineno)d)"

_END_SYMBOLS = (".", "!", "?", "。", "！", "？")
_FINAL_STYLES = {"bold", "selected"}

_TB_STR = "traceback"
_EXC_ARROW_STR = cstr("\u25B6 ", styles={"bold"})
_CAUSE_LINE = {"vertical": "\u2502", "horizontal": "\u2500", "corner": "\u2514"}
# traceback braces
_BRACE = {"upper": "\u250E", "middle": "\u2520", "lower": "\u2516"}
_SHORT_LINE = "\u2574"

_FRAME = {
    "light": {
        "top_left": "\u256D",
        "top_right": "\u256E",
        "bottom_left": "\u2570",
        "bottom_right": "\u256F",
        "horizontal": "\u2500",
        "vertical": "\u2502"
    },
    "double": {
        "top_left": "\u2554",
        "top_right": "\u2557",
        "bottom_left": "\u255A",
        "bottom_right": "\u255D",
        "horizontal": "\u2550",
        "vertical": "\u2551"
    },
    "heavy": {
        "top_left": "\u250F",
        "top_right": "\u2513",
        "bottom_left": "\u2517",
        "bottom_right": "\u251B",
        "horizontal": "\u2501",
        "vertical": "\u2503"
    }
}


def box_lines(lines: Sequence[str], /, top_indent: int = 0, rest_indent: int = 0, frame_style: str = "light", **pattern: Any) -> str:
    r"""
    Box the lines with a border.

    Parameters
    ----------
        lines : Sequence[str]
            The lines to be boxed.

        top_indent : int, default to `0`
            The number of spaces to indent for the top border line.

        rest_indent : int, default to `0`
            The number of spaces to indent for the rest lines.

        frame_style : str, default to `"light"`
            The border style name of :attr:`_FRAME`, including `"light"`, `"double"` and `"heavy"`.

        **pattern : Any
            The color pattern for the border, including `fg`, `bg` and `styles`.
            See also :func:`cobra_color.cstr` for details.

    Returns
    -------
        str
            The boxed lines as a string.
    """
    frame = _FRAME[frame_style]
    rest_indent = max(0, rest_indent)
    # width
    width = max(len(line) for line in lines) + 1
    if width == 1:
        return ""
    width = max(_TRACE_CONFIG.get("min_width", 50), width)
    # top line
    top_line_items: List[str] = []
    top_indent = max(0, top_indent)
    if top_indent == 0:
        top_line_items.append(frame["top_left"])
    top_remain_len = width + rest_indent + 1 - top_indent
    if top_remain_len > 0:
        top_line_items.append(frame["horizontal"] * top_remain_len)
    if top_remain_len > -1:
        top_line_items.append(frame["top_right"])
    boxed_lines = [cstr(*top_line_items, **pattern)]
    # middle lines
    vert_line = cstr(frame["vertical"], **pattern)
    for line in lines:
        boxed_lines.append(cstr(" " * rest_indent, vert_line, line, " " * (width - len(line)), vert_line))
    # bottom line
    boxed_lines.append(cstr(
        " " * rest_indent,
        frame["bottom_left"],
        frame["horizontal"] * width,
        frame["bottom_right"],
        **pattern
    ))
    return cstr(*boxed_lines, sep="\n")


def trace_stack(stack_level: int = 1, /, fmt: str = _TRACE_FMT) -> str:
    r"""
    Trace the stack information of the function call.

    Parameters
    ----------
        stack_level : int, default to `1`
            The stack level of the function call.
            - `0`: this function;
            - `1`: the caller of this function;
            - ...

        fmt : str, default to TRACE_FMT
            The format string, :attr:`_TRACE_FMT` is defined as `r"%(fileName)s->%(funcName)s(%(lineno)d)"`.
            Includes:
            - `%(stackDepth)d`: stack depth;
            - `%(funcName)s`: function name;
            - `%(filePath)s`: file path;
            - `%(fileName)s`: file name;
            - `%(lineno)d`: line number;
            - `%(funcLineno)d`: function start line number.

    Returns
    -------
        str
            Formatted stack information.
    """
    total_stack = inspect.stack()
    stack_level = min(max(0, stack_level), len(total_stack) - 1)
    frame_info = total_stack[stack_level][0]
    # format
    args = {}
    if r"%(stackDepth)d" in fmt:  # current stack depth
        args["stackDepth"] = len(total_stack) - stack_level
    if r"%(funcName)s" in fmt:  # current function name
        args["funcName"] = frame_info.f_code.co_name
    if r"%(filePath)s" in fmt:  # current file path
        args["filePath"] = frame_info.f_code.co_filename
    if r"%(fileName)s" in fmt:  # current file name
        args["fileName"] = os.path.basename(frame_info.f_code.co_filename)
    if r"%(lineno)d" in fmt:  # current line number
        args["lineno"] = int(frame_info.f_lineno)
    if r"%(funcLineno)d" in fmt:  # function start line number
        args["funcLineno"] = frame_info.f_code.co_firstlineno

    return fmt % args


def trace_exc(
    exception: Any,
    /,
    mode: Literal["cause", "context"] = "context",
    exc_depth: int = -1,
    tb_depth: Optional[int] = -1,
    exc_args_limit: int = -1,
    indent: int = 4
) -> str:
    r"""
    Trace the exception chain and format it as a string.

    Parameters
    ----------
        exception : Any
            The exception instance to trace.

        mode : Literal["cause", "context"], default to `"context"`
            The trace mode for exception chain.
            - `"cause"`: Trace the cause exception chain, i.e., the exceptions linked by `__cause__` attribute. This chain is formed when exceptions are raised with `from` keyword.
            - `"context"`: Trace the context exception chain, i.e., the exceptions linked by `__context__` attribute. This chain is formed when an exception is raised in an except block and not handled, regardless of whether `from` keyword is used.

        exc_depth : int, default to `-1`
            The maximum depth of exception chain to trace.
            - `< 1`: Trace all exceptions in the chain;
            - `>= 1`: Trace the last :param:`exc_depth` exceptions in the chain, omitting earlier exceptions.

        tb_depth : Optional[int], default to `-1`
            The maximum depth of stack traceback frames to trace for each exception.
            - `None`: Without traceback frames;
            - `< 1`: Trace all traceback frames;
            - `>= 1`: Trace the last :param:`tb_depth` traceback frames, omitting earlier frames.

        exc_args_limit : int, default to `-1`
            The maximum number of exception arguments to display for each exception.
            - `< 1`: Display all exception arguments;
            - `>= 1`: Display the first :param:`exc_args_limit` exception arguments, omitting later arguments.

        indent : int, default to `4`
            The number of spaces to indent for each level of exception in the chain.

    Returns
    -------
        str
            Formatted string.
    """
    _base_indent = _add_indent(indent)

    # === non-exception type ===
    if not isinstance(exception, Exception):
        # non-exception type
        return _base_indent + _EXC_ARROW_STR +\
            cstr(_fmt_msg(exception), styles=_FINAL_STYLES)

    # === exception type ===
    trace_lines: List[str] = []
    # 1. build exception chain
    exc_lists: List[List[Any]] = [[]]
    exc = exception
    _exc_list = exc_lists[-1]
    exc_count = 0
    while exc is not None:
        _exc_list.append((exc_count, exc))
        exc_count += 1
        # check next exception in chain
        if exc.__cause__ is not None:
            exc = exc.__cause__
        else:
            if mode == "cause":
                # cause
                break
            else:
                # context
                exc_lists.append([])
                _exc_list = exc_lists[-1]
                exc = exc.__context__

    if mode == "cause":
        trace_lines.extend([
            _base_indent,
            cstr("    for cause exception(s) only", styles={"dim"})
        ])
    if 0 < exc_depth < exc_count:
        _e_omit_num = _rest_omit_num = exc_count - exc_depth
        while _rest_omit_num > 0:
            if _rest_omit_num >= len(exc_lists[0]):
                _rest_omit_num -= len(exc_lists[0])
                exc_lists.pop(0)
            else:
                exc_lists[0] = exc_lists[0][_rest_omit_num:]
                _rest_omit_num = 0
        trace_lines.extend([
            _base_indent,
            cstr(f"... omitted {_e_omit_num} exception(s) ...", styles={"dim"})
        ])
    # 2. format exception chain
    for _exc_list in exc_lists:
        trace_lines.extend(_compile_trace(
            _exc_list,
            total_exc=exc_count,
            tb_depth=tb_depth,
            exc_args_limit=exc_args_limit,
            indent=indent
        ))

    return cstr(*trace_lines)


def _compile_trace(
    exc_cause_list: List[Tuple[int, Exception]],
    /,
    total_exc: int,
    tb_depth: Optional[int],
    exc_args_limit: int,
    indent: int
):
    r"""Compile the exception chain into trace lines."""
    trace_lines: List[str] = []
    for idx, (exc_idx, exc) in enumerate(exc_cause_list):
        # for exception chain
        _base_indent = _add_indent(indent)
        is_final = exc_idx == total_exc - 1
        # actual exception
        if len(exc_cause_list) > 1 and idx > 0:
            # non-first exception in chain with multiple exceptions
            indent_arrow = cstr(
                _add_indent(indent-3),
                _CAUSE_LINE["corner"] + "<" + _CAUSE_LINE["horizontal"],
                f"({total_exc - exc_idx})",
                _EXC_ARROW_STR
            )
        else:
            indent_arrow = cstr(_base_indent, f"({total_exc - exc_idx})", _EXC_ARROW_STR)
        trace_lines.append(indent_arrow)
        # exception name
        exc_name = cstr(
            exc.__class__.__name__,
            fg=None if is_final else "c",
            bg="lr" if is_final else None,
            styles={"dim"} if is_final else {"bold"}
        )
        # 3. stack traceback
        if len(exc_cause_list) > 1 and idx < len(exc_cause_list) - 1:
            # non-final exception in chain with multiple exceptions
            _tb_indent = cstr(_base_indent, f" {_CAUSE_LINE['vertical']}   ")
        else:
            _tb_indent = _add_indent(len(indent_arrow) - 1)
        tb_stack = list(traceback.extract_tb(exc.__traceback__))
        if not tb_depth or len(tb_stack) == 1:
            # only one traceback frame
            tb_stack = tb_stack[-1:]
            _name_width = len(exc_name)
        else:
            if 0 < tb_depth < len(tb_stack):
                _tb_omit_num = len(tb_stack) - tb_depth
                tb_stack = tb_stack[-tb_depth:]
                _tb_len = len(tb_stack)
                tb_stack.insert(0, cstr(f"... omitted {_tb_omit_num} traceback frame(s) ...", styles={"dim"}))
            _tb_len = len(tb_stack)
            tb_idx_width = len(str(_tb_len))
            _name_width = max(len(_TB_STR) + tb_idx_width + 1, len(exc_name))
        for tb_idx, tb in enumerate(tb_stack):
            # for each traceback frame
            if isinstance(tb, str):
                # omitted traceback frames
                trace_lines.extend([f"{_BRACE['upper']} ", tb])
                continue
            # traceback detail
            _tb_detail = cstr(f"{tb.filename}({tb.lineno})", styles={"udl"})
            if tb_idx == len(tb_stack) - 1:
                # final traceback (exception raised here)
                tb_prefix = _BRACE["lower"] if len(tb_stack) > 1 else ""
                _tb_name = exc_name
                _tb_styles = _FINAL_STYLES if is_final else set()
                # 4. exception args
                _e_args = _fmt_exc_args(exc.args, is_final)
                if 0 < exc_args_limit < len(_e_args):
                    _arg_omit_num = len(_e_args) - exc_args_limit
                    _e_args = _e_args[:exc_args_limit]
                    _e_args.append(
                        cstr(f" ... omitted {_arg_omit_num} argument(s) ...", styles={"dim"})
                    )
                _e_args_space = _add_indent(_name_width + len(_tb_detail) + len(tb_prefix) + 5, nowrap=True)
                arg_lines = []
                for e_arg_idx, e_arg in enumerate(_e_args):
                    # for each exception argument
                    if e_arg_idx == 0:
                        # first exception argument
                        _prefix = f" {_BRACE['upper']}" if len(_e_args) > 1 else " "
                        arg_lines.append(cstr(_prefix, e_arg, styles=[(None, _tb_styles)]))
                    else:
                        # non-first exception argument
                        _prefix = _BRACE["lower"] if e_arg_idx == len(_e_args) - 1 else _BRACE["middle"]
                        arg_lines.extend([_tb_indent, _e_args_space, _prefix, e_arg])
            else:
                # non-final traceback
                tb_prefix = _BRACE["upper"] if tb_idx == 0 else _BRACE["middle"]
                _tb_name = cstr(
                    f"{_TB_STR}-{str(len(tb_stack) - tb_idx - 1).zfill(tb_idx_width)}",
                    fg="lg",
                    styles={"bold"}
                )
                _tb_styles = {"dim"}
                arg_lines = [cstr(" >>> ", fg="ly", styles={"bold", *_tb_styles}),
                             cstr(tb.line, styles=_tb_styles)]
            tb_indent = "" if tb_idx == 0 else _tb_indent
            # traceback head
            tb_head = cstr("<", _tb_name.center(_name_width), ": ", _tb_detail,
                           ">", styles=[(None, {"italic", *_tb_styles})])
            # 5. assemble
            trace_lines.extend([tb_indent, tb_prefix, tb_head, *arg_lines])
        # update
        indent += 4

    return trace_lines


def _fmt_msg(msg: Any, /, limit: Optional[int] = None) -> str:
    r"""Format message."""
    msg = msg if isinstance(msg, str) else str(msg)
    if limit is not None and len(msg) > limit:
        msg = msg[:limit] + ".."
    if msg and not msg.endswith(_END_SYMBOLS):
        msg += "."
    return msg


def _add_indent(indent: int, /, nowrap: bool = False) -> str:
    r"""Create indentation."""
    return cstr(("" if nowrap else "\n") + " " * indent, styles={"disappear"})


def _fmt_exc_args(e_args: Tuple[Any, ...], /, is_final: bool, is_top=True) -> List[str]:
    r"""Format exception arguments."""
    fmt_args: List[str] = []
    for m in e_args:
        if isinstance(m, Tuple):
            fmt_args.extend(_fmt_exc_args(m, is_final, is_top=is_top))
        elif isinstance(m, Exception):
            # exception head
            exc_name = cstr(m.__class__.__name__, fg="r" if is_final else "c")
            e_msg = cstr("<", exc_name, "> ", styles={"italic"})
            # exception args
            sub_args = _fmt_exc_args(m.args, is_final, is_top=False)
            if len(sub_args) > 1:
                rest_msgs = [f", {_fmt_msg(arg, 12)}" for arg in sub_args[1:]]
                e_msg += cstr("( ", styles={"dim"}) + sub_args[0] +\
                    cstr(*rest_msgs, " )", styles={"dim"})
            else:
                e_msg += sub_args[0]
            fmt_args.append(e_msg)
        else:
            prefix = _SHORT_LINE if is_top and len(e_args) > 1 else ""
            fmt_args.append(prefix + _fmt_msg(m))

    return fmt_args



# src/cobra_log/__init__.py
"""
cobra-log
===========

A lightweight and easy-to-use logging library for Python.

Functions
---------
- :func:`log_init`: Initialize the logging system.
- :func:`enable_color`: Enable or disable colored output (need :pkg:`cobra-color`).
- :func:`set_trace`: Configure global trace display settings.
- :func:`critical`: Log a critical error message and raise an exception.
- :func:`error`: Log an error message.
- :func:`warning`: Log a warning message.
- :func:`info`: Log an informational message.
- :func:`debug`: Log a debug message.
- :func:`trace_stack`: Trace the stack information of the function call.

Examples
--------

```python
from cobra_log import (log_init, info, warning, error)

# Initialize the log system
log_init("log_save_path.log", use_color=True)

try:
    try:
        try:
            1 / 0
        except Exception as e:
            raise error("An error occurred.", throw=None) from e
    except Exception as ee:
        raise error(
            "An error(TimeoutError) occurred.",
            throw=TimeoutError
        ) from ee
except Exception:
    critical("A critical error occurred.", throw=None)
```
"""

from ._core import (log_init, enable_color, set_trace)
from ._log_levels import (critical, error, warning, info, debug)
from ._utils import trace_stack


__author__ = "Zhen Tian"
__version__ = "1.2.0"

__all__ = [
    "log_init",
    "enable_color",
    "set_trace",
    "critical",
    "error",
    "warning",
    "info",
    "debug",
    "trace_stack"
]



# -*- coding: utf-8 -*-
# Python version: 3.9
# @TianZhen

from __future__ import annotations
import logging
from logging.handlers import RotatingFileHandler
import warnings
import os
from typing import (Any, Optional)

from .types import LogLevelName


_LOG_FMT = r"%(asctime)s - <%(levelname)s> - <%(filename)s(%(funcName)s)-%(lineno)d> - %(message)s"
_DATE_FMT = r"%y-%m-%d %H:%M:%S"

# Logger instance
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.DEBUG)
_FILE_HANDLER = None
# cstr & smart_print
try:
    from cobra_color import (cstr as t_cstr, safe_print as t_safe_print)  # type: ignore
    _COLOR_AVAIL = True
except ImportError:
    warnings.warn(
        "Missing color library `cobra-color`, terminal color display unavailable for `cobra-log`.",
        category=UserWarning,
        stacklevel=3
    )
    _COLOR_AVAIL = False

_USE_COLOR: bool = _COLOR_AVAIL


def cstr(*objects: Any, sep: str = "", **kwargs):
    if _USE_COLOR:
        return t_cstr(*objects, sep=sep, **kwargs)
    return sep.join(objects)


def display(*args, **kwargs):
    if _COLOR_AVAIL:
        return t_safe_print(*args, **kwargs)
    print(*args, **kwargs)


def enable_color(flag: bool = True, /):
    r"""
    Enable or disable colored terminal display. Requires :pkg:`cobra-color`.
    """
    global _USE_COLOR
    _USE_COLOR = flag and _COLOR_AVAIL


# trace display
_TRACE_CONFIG = {
    "with_border": True,
    "exc_mode": "context",
    "exc_depth": -1,
    "tb_depth": -1,
    "exc_args_limit": -1,
    "min_width": 50
}


def set_trace(**kwargs: Any):
    r"""
    Set the global trace display configuration.

    Parameters
    ----------
        **kwargs : Any
            The trace display configuration to be updated. Including:
            - `with_border`: bool
            - `exc_mode`: Literal["cause", "context"]
            - `exc_depth`: int
            - `tb_depth`: Optional[int]
            - `exc_args_limit`: int
    """
    global _TRACE_CONFIG
    _TRACE_CONFIG.update(kwargs)


def log_init(
    log_save_path: Optional[str] = None,
    log_level: LogLevelName = "debug",
    log_fmt: str = _LOG_FMT,
    date_fmt: str = _DATE_FMT,
    cover: bool = True,
    backup_count: int = 0,
    use_color: bool = True
):
    r"""
    Initialize the log storage configuration.

    NOTE: The log storage configuration can be set multiple times, and only the latest log storage configuration instance is retained.

    Parameters
    ----------
        log_save_path : Optional[str], default to `None`
            The log file storage address. If is `None`, no log file output stream is created.

        log_level : LogLevelName, default to `"debug"`
            The lowest level of log file storage, :type:`LogLevelName` includes `"debug"`, `"info"`, `"warning"`, `"error"` and `"critical"`.

        log_fmt : str, default to LOG_FMT
            The log storage format, :attr:`_LOG_FMT` is defined as `r"%(asctime)s - <%(levelname)s> - <%(filename)s(%(funcName)s)-%(lineno)d> - %(message)s"`.
            Includes:
            - `%(levelno)s`: log level value;
            - `%(levelname)s`: log level name;
            - `%(pathname)s`: the path of the current executable program, i.e., sys.argv[0];
            - `%(filename)s`: the name of the current executable program;
            - `%(funcName)s`: the current function of the log;
            - `%(lineno)d`: the current line number of the log;
            - `%(asctime)s`: the time of the log;
            - `%(thread)d`: thread ID;
            - `%(threadName)s`: thread name;
            - `%(process)d`: process ID;
            - `%(message)s`: log information;
            - `%(name)s`: log handler name, default to root.

        date_fmt : _type_, default to DATE_FMT
            The time storage format in the log, :attr:`_DATE_FMT` is defined as `r"%y-%m-%d %H:%M:%S"`.

        cover : bool, default to `True`
            Control whether to cover the latest storage configuration instance.

        backup_count : int, default to `0`
            The number of log file backups.
            - `>0`: rolling log;
            - `<=0`: no backup, all logs are stored in one file.

        use_color : bool, default to `True`
            Control whether to enable colored terminal display. Requires :pkg:`cobra-color`.
    """
    # log level
    if log_level == "critical":
        level = logging.CRITICAL
    elif log_level == "error":
        level = logging.ERROR
    elif log_level == "warning":
        level = logging.WARNING
    elif log_level == "info":
        level = logging.INFO
    else:
        level = logging.DEBUG
    # set logger handler
    global _FILE_HANDLER
    if _FILE_HANDLER and cover:
        _LOGGER.removeHandler(_FILE_HANDLER)

    if log_save_path is not None and log_save_path.endswith(".log"):
        # create log directory
        log_dir = os.path.dirname(log_save_path)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
        # file handler
        _FILE_HANDLER = RotatingFileHandler(
            filename=log_save_path,
            mode="a",
            backupCount=backup_count,
            maxBytes=100*1024,
            delay=True
        )
        _LOGGER.addHandler(_FILE_HANDLER)
        _FILE_HANDLER.setFormatter(logging.Formatter(
            fmt=log_fmt,
            datefmt=date_fmt
        ))
        _FILE_HANDLER.setLevel(level)

    # enable color
    enable_color(use_color)
